package dev.pfaff.altargui.widgets;

import dev.pfaff.altargui.*;
import dev.pfaff.altargui.internal.util.StencilLayoutContext;
import dev.pfaff.altargui.types.Alignment;
import org.jetbrains.annotations.NotNull;

public sealed interface PositionedWidget extends SingleChildWidget {
	public record Absolute(float x, float y, Widget child) implements PositionedWidget {
		@Override
		public @NotNull WidgetInstance<Absolute> instantiate(WidgetInstantiationArgs args) {
			return new State(args);
		}

		private static final class State extends CustomLayoutState<Absolute> {
			private float x;
			private float y;

			private State(WidgetInstantiationArgs args) {
				super(args);
			}

			@Override
			public void configure(StateContext stateContext, Absolute widget, boolean init) {
				super.configure(stateContext, widget, init);
				if (x != widget.x || y != widget.y) {
					x = widget.x;
					y = widget.y;
					markNeedsLayout();
				}
			}

			@Override
			protected void computeStencilLayout(StateContext context) {
				StencilLayoutContext.updateMountAndLayout(context, this, getChildElement(), (element, ctx) -> {
//					var allocWidth = mcWindow.getScaledWidth();
//					var allocHeight = mcWindow.getScaledHeight();
					var allocWidth = this.allocatedWidth();
					var allocHeight = this.allocatedHeight();
					element.setAvailableSize(allocWidth - this.x, allocHeight - this.y);
					element.stencilLayout(ctx);
					setComputedSize(0, 0);
				});
			}

			@Override
			protected void computeTraceLayout() {
				var e = getChildElement();
				e.setPosition(this.x, this.y);
				e.traceLayout();
			}
		}
	}

	public record Relative(float x, float y, Widget child) implements PositionedWidget {
		@Override
		public @NotNull WidgetInstance<Relative> instantiate(WidgetInstantiationArgs args) {
			return new State(args);
		}

		private static final class State extends CustomLayoutState<Relative> {
			private float x;
			private float y;

			private State(WidgetInstantiationArgs args) {
				super(args);
			}

			@Override
			public void configure(StateContext stateContext, Relative widget, boolean init) {
				super.configure(stateContext, widget, init);
				if (x != widget.x || y != widget.y) {
					x = widget.x;
					y = widget.y;
					markNeedsLayout();
				}
			}

			@Override
			protected void computeStencilLayout(StateContext context) {
				StencilLayoutContext.updateMountAndLayout(context, this, getChildElement(), (element, ctx) -> {
					element.setAvailableSize(allocatedWidth() - x, allocatedHeight() - y);
					element.stencilLayout(ctx);
					setComputedSize(Math.max(0f, x + element.computedWidth()), Math.max(0f, y + element.computedHeight()));
				});
			}

			@Override
			protected void computeTraceLayout() {
				var e = getChildElement();
				e.setPosition(allocatedX() + x, allocatedY() + y);
				e.traceLayout();
			}
		}
	}

	public record Aligned(Widget child, Alignment horizontal, Alignment vertical, boolean clamp) implements PositionedWidget {
		public Aligned(Widget child, Alignment horizontal, Alignment vertical) {
			this(child, horizontal, vertical, true);
		}

		@Override
		public @NotNull WidgetInstance<Aligned> instantiate(WidgetInstantiationArgs args) {
			return new State(args);
		}

		private static final class State extends CustomLayoutState<Aligned> {
			private @NotNull Alignment horizontal = Alignment.Start;
			private @NotNull Alignment vertical = Alignment.Start;
			private boolean clamp;

			private float childOffsetX;
			private float childOffsetY;

			private State(WidgetInstantiationArgs args) {
				super(args);
			}

			@Override
			public void configure(StateContext stateContext, Aligned widget, boolean init) {
				super.configure(stateContext, widget, init);
				var dirtyStencil = false;
				if (horizontal != widget.horizontal) {
					horizontal = widget.horizontal;
					dirtyStencil = true;
				}
				if (vertical != widget.vertical) {
					vertical = widget.vertical;
					dirtyStencil = true;
				}
				if (clamp != widget.clamp) {
					clamp = widget.clamp;
					dirtyStencil = true;
				}

				if (dirtyStencil) {
					markNeedsStencilLayout();
				}
			}

			@Override
			protected void computeStencilLayout(StateContext context) {
				var horizontal = this.horizontal;
				var vertical = this.vertical;
				StencilLayoutContext.updateMountAndLayout(context, this, getChildElement(), (e, ctx) -> {
					e.copyAvailableSizeFrom(this);
					e.stencilLayout(ctx);
					var allocWidth = allocatedWidth();
					var allocHeight = allocatedHeight();
					var childOffsetX = horizontal.apply(e.computedWidth(), allocWidth);
					var childOffsetY = vertical.apply(e.computedHeight(), allocHeight);
					if (this.clamp) {
						childOffsetX = Math.max(childOffsetX, 0.0f);
						childOffsetY = Math.max(childOffsetY, 0.0f);
					}

					var dirtyTrace = false;
					if (this.childOffsetX != childOffsetX) {
						this.childOffsetX = childOffsetX;
						dirtyTrace = true;
					}
					if (this.childOffsetY != childOffsetY) {
						this.childOffsetY = childOffsetY;
						dirtyTrace = true;
					}
					if (dirtyTrace) {
						markNeedsTraceLayout();
					}
					var clamp = this.clamp;
					setComputedSize(switch (horizontal) {
						case Start -> childOffsetX + e.computedWidth();
						case Center, End -> clamp ? Math.max(allocWidth, e.computedWidth()) : allocWidth;
					}, switch (vertical) {
						case Start -> childOffsetY + e.computedHeight();
						case Center, End -> clamp ? Math.max(allocHeight, e.computedHeight()) : allocHeight;
					});
				});
			}

			@Override
			protected void computeTraceLayout() {
				var e = getChildElement();
				e.setPosition(allocatedX() + this.childOffsetX, allocatedY() + this.childOffsetY);
				e.traceLayout();
			}
		}
	}
}
